import json
import os
import re
import sys

import tabulate

from .module_manager import ModuleManager
from .utils.exp_util import get_relative_imports
from .utils.git_util import parse_url_from_git
from ...mate_config import MateConfig
from .sources.remote import RemoteDataSource
from .sources.local.local import LocalDataSource
from pipreqs import pipreqs

from .package import Package
from typing import Optional
import ipdb


class ModuleRepository:
    def __init__(self, config: MateConfig, run_local_api_server: bool = False):
        self.config = config
        self.package_manager = ModuleManager(config)
        self.remote = RemoteDataSource()
        self.local = LocalDataSource(config)
        # self.__generate_pip_requirements(self.config.project)

    @staticmethod
    def init_project(project_name: str):

        if not os.path.exists(project_name):
            os.mkdir(project_name)
            os.chdir(project_name)
        else:
            print("Project already exists")
            sys.exit(1)

        mate_json = os.path.join("mate.json")
        if not os.path.exists(mate_json):
            dic = {
                "project": project_name,
            }
            # create mate.json

            with open(mate_json, "w") as f:
                json.dump(dic, f, indent=4)
        else:
            print("Project already exists")
            sys.exit(1)

        if not os.path.exists(project_name):
            os.mkdir(project_name)
            init__file = os.path.join(project_name, "__init__.py")
            open(init__file, "a").close()
        try:
            folders = ["experiments", "models", "data", "trainers"]
            for folder in folders:
                os.makedirs(os.path.join(project_name, folder), exist_ok=True)
                init__file = os.path.join(project_name, folder, "__init__.py")
                if not os.path.exists(init__file):
                    open(init__file, "a").close()
            print(
                "Project {} created, run `cd {}` to enter the project folder".format(
                    project_name, project_name
                )
            )

        except Exception as e:
            print(e)

    def install_url(self, url: str, *args, **kwargs):

        self.package_manager.install_package(url, *args, **kwargs)

    def auto(self, command: str, *args):
        if command == "export":
            self.__export()
        elif command in ["init", "fix", "i"]:
            self.__generate__init__(self.config.project)

    def __generate__init__(self, root: str = None):
        init__py = os.path.join(root, "__init__.py")
        if not os.path.exists(init__py):
            with open(init__py, "w") as f:
                f.write("")
            print(f"Created {init__py}")

        for folder in os.listdir(root):
            path = os.path.join(root, folder)
            if not os.path.isdir(path) or folder == "__pycache__" or "." in folder:
                continue
            init__py = os.path.join(path, "__init__.py")
            if not os.path.exists(init__py):
                with open(init__py, "w") as f:
                    f.write("")
                print(f"Created {init__py}")
            self.__generate__init__(path)

    def __parse_index_urls(self, reqs: list[str]):
        urls = {
            "torch": "https://download.pytorch.org/whl/torch_stable.html",
            "jax": "https://storage.googleapis.com/jax-releases/jax_releases.html",
        }
        indexes = set()
        for req in reqs:
            if "torch" in req:
                indexes.add(urls["torch"])
            if "jax" in req:
                indexes.add(urls["jax"])

        return indexes

    def __add_index_url_to_requirements(self, path: str):
        with open(os.path.join(path), "r") as f:
            lines = f.readlines()
        linecount = len(lines)
        lines = [
            line
            for line in lines
            if not ".egg>=info" in line
            and not ".egg==info" in line
            and not ".egg~=info" in line
        ]

        # remove +cu{numbers} version form lines
        # regex for numbers with at least 1 digit
        regex = re.compile(r"\+cu\d+")
        lines = [regex.sub("", line) for line in lines]

        urls = self.__parse_index_urls(lines)
        if len(urls) > 0:
            with open(os.path.join(path), "w") as f:
                for url in urls:
                    f.write(f"--extra-index-url {url}\n")
                for line in lines:
                    f.write(line)
        elif linecount != len(lines):
            with open(os.path.join(path), "w") as f:
                for line in lines:
                    f.write(line)

    def __generate_deps_in_depth(self, root_path):

        # init__path = os.path.join(path, "__init__.py")

        for dir in os.listdir(root_path):
            if dir.startswith(".") or dir.startswith("__"):
                continue
            path = os.path.join(root_path, dir)
            if os.path.isdir(path):
                # check if this is a python module
                init__path = os.path.join(root_path, dir, "__init__.py")
                if not os.path.exists(init__path):
                    continue
                # if dir in ["trainers", "experiments", "models", "data"] and
                if not (
                    dir in ["trainers", "experiments", "models", "data"]
                    and self.config.project in root_path
                ):
                    self.__generate_pip_requirements(path)

                self.__generate_deps_in_depth(path)

    def __export(self):
        self.__generate_sub_pip_reqs()

        modules = self.list()

        table = []

        for key, value in modules.items():
            if type(value) is list:
                table.append([{"type": key, "name": name} for name in value])
            elif type(value) is dict:
                table.append([{"type": key, "name": name} for name in value.keys()])

        # ipdb.set_trace()

        table = [item for sublist in table for item in sublist]

        # add url to each item in table

        base_url = parse_url_from_git()
        user_name = base_url.split("/")[3]
        repo_name = base_url.split("/")[4]
        for item in table:

            item[
                "url"
            ] = f"{base_url}{self.config.project}/{item['type']}/{item['name']}"
            item[
                "short_url"
            ] = f"{user_name}/{repo_name}/{self.config.project}/{item['type']}/{item['name']}"

            # if repo name is same as project name
            if repo_name == self.config.project:
                # item["url"] = f"{base_url}/{item['type']}/{item['name']}"
                # get user name from url
                item[
                    "short_url"
                ] = f"{user_name}/{repo_name}/{item['type']}/{item['name']}"

        # read dependencies
        for item in table:
            path = os.path.join(
                self.config.project, item["type"], item["name"], "requirements.txt"
            )
            dep_path = os.path.join(
                self.config.project, item["type"], item["name"], "dependencies.json"
            )

            if os.path.exists(path):
                with open(path, "r") as f:
                    item["dependencies"] = f.readlines()
            if os.path.exists(dep_path):
                with open(dep_path, "r") as f:
                    if "dependencies" in item:
                        item["dependencies"] += json.load(f)["dependencies"]
                    else:
                        item["dependencies"] = json.load(f)["dependencies"]
                    # item["module_dependencies"] = json.load(f)

            if "dependencies" in item:
                item["dependencies"] = [
                    dep.replace("\n", "") for dep in item["dependencies"]
                ]

        table = tabulate.tabulate(
            table,
            headers="keys",
            tablefmt="github",
            showindex="always",
            disable_numparse=True,
        )

        # save table to export.md
        with open("export.md", "w") as f:
            f.write(table)

        print("Exported to export.md")

    def __generate_sub_pip_reqs(self):

        root_path = self.config.project
        self.__generate_deps_in_depth(root_path)

        for dir in os.listdir("."):
            if (
                dir.startswith(".")
                or dir.startswith("__")
                or dir == self.config.project
            ):
                continue
            path = os.path.join(".", dir)
            if os.path.isdir(path):
                # check if this is a python module
                init__path = os.path.join(".", dir, "__init__.py")
                if not os.path.exists(init__path):
                    continue
                self.__generate_pip_requirements(path)

    def __generate_mate_dependencies(self, path):
        # ipdb.set_trace()

        files = [f for f in os.listdir(path) if f.endswith(".py") and "__" not in f]
        original_files = [file.replace(".py", "") for file in files] + [
            f for f in os.listdir(path) if "__" not in f
        ]

        relative_imports = [get_relative_imports(os.path.join(path, f)) for f in files]
        # flatten array to unique set
        relative_imports = set(
            [item for sublist in relative_imports for item in sublist]
        )

        relative_imports = [
            module
            for module in relative_imports
            if not any([file in module for file in original_files])
        ]

        url_git = parse_url_from_git()

        if url_git is None:
            print("No git url found, skipping dependencies.json")
            return

        deps = set()
        for module in relative_imports:

            if module.endswith(".py"):
                continue
            # if its a python file, return parent module

            tpath = [self.config.project, *module.split(".")]
            tpath[-1] = tpath[-1] + ".py"

            sister_module_path = [*module.split(".")]

            if os.path.exists(os.path.join(*tpath)):
                # module = parent
                url = "/".join(tpath[:-1])
            elif os.path.exists(os.path.join(*sister_module_path)):
                url = sister_module_path[0] + "/"
            else:
                url = self.config.project + "/" + module.replace(".", "/")

            if url_git:
                url = url_git + url
            deps.add(url)

        if len(deps) == 0:
            return

        try:

            deps_json = os.path.join(path, "dependencies.json")
            if os.path.exists(deps_json):
                with open(deps_json, "r") as f:
                    # ipdb.set_trace()
                    deps_json = json.load(f)
                    if "env" in deps_json:
                        env = deps_json["env"]
                    else:
                        env = {}

            else:
                env = {}
        except Exception as e:
            print(f"Error reading {path}/dependencies.json, skipping env")
            env = {}

        with open(os.path.join(path, "dependencies.json"), "w") as f:
            deps = {"dependencies": list(deps), "env": env}
            json.dump(deps, f, indent=4)
            print(f"Generated dependencies.json for {path}")

    def __generate_pip_requirements(self, path):

        # ipdb.set_trace()
        imports = pipreqs.get_all_imports(path)
        # import_info_remote = pipreqs.get_imports_info(imports)

        import_info_local = pipreqs.get_import_local(imports)

        self.__generate_mate_dependencies(path)

        import_info = []

        if path == self.config.project:
            pipreqs.generate_requirements_file(
                "requirements.txt", import_info_local, "~="
            )
            self.__add_index_url_to_requirements("requirements.txt")

        else:
            pipreqs.generate_requirements_file(
                os.path.join(path, "requirements.txt"), import_info_local, "~="
            )
            self.__add_index_url_to_requirements(os.path.join(path, "requirements.txt"))
            print(f"Generated requirements.txt for {path}")

        for im in import_info_local:

            name = im["name"]
            version = im["version"]

            res = {
                "name": name,
                "version": version,
            }

            import_info.append(res)

        return {"pip": import_info}

    def list(self, module: str = None):
        if module == None:
            return self.local.summary()
        return self.local.list(module)

    def get_mate_summary(self):
        return self.local.summary()

    def install_package(self, package: Package):
        self.local.install_package(package)



"""
## Cli Parser

Mate's cli parser is a simple parser that parses the command line arguments and calls the appropriate method on the Mate class.

Notice that for boolean arguments, you can use either false or False, true or True. And for None you can use either null or None.

**Example**

```
mate init my_project venv=false
```

"""
from .mate_config import MateConfig
from .utils import print_markdown, remove_indent
from .mate import Mate
import inspect
from typing import Optional, Callable, Sized
import ipdb
import os
import sys


def __parse_signature(class_name, method_name: str):
    return tuple(
        val
        for (name, val) in inspect.signature(
            getattr(class_name, method_name)
        ).parameters.items()
        if name != "self"
    )


def __get_methods_with_arguments(class_name):
    return {
        k: __parse_signature(class_name, k)
        for k in class_name.__dict__.keys()
        if not k.startswith("_")
    }


def method_to_md(method_name, member: Optional[Callable] = None):
    if member is None:
        member = getattr(MateCLI, method_name)
    assert member is not None
    params = inspect.signature(member).parameters.values()
    inline_params = " ".join(
        [
            f"<{param.name}{('=' + str(param.default)) if (param.default != inspect._empty) else ''}>"
            for param in params
            if param.name != "self"
        ]
    )
    param_descriptions = {
        line[1].split(" ")[1]: ":".join(line[2:])
        for line in [l.strip().split(":") for l in str(member.__doc__).split("\n")]
        if len(line) > 1 and line[1].startswith("param")
    }

    raw_list_params = [
        f"- {param.name} : `{param.annotation.__name__}` : {param_descriptions.get(param.name, '')}"
        + (f"={param.default}" if param.default != inspect._empty else "")
        for param in params
        if param.name != "self"
    ]
    list_params = "\n".join(raw_list_params)
    method_description = "\n".join(
        [
            line
            for line in str(member.__doc__).split("\n")
            if not line.strip().startswith(":param")
        ]
    )

    code = remove_indent(
        f"""
    ```
      ❯ mate {method_name} {inline_params}
    ```
    """
    ).strip()
    doc = remove_indent(
        f"""
{code}

**Params**

{list_params.strip()}

{remove_indent(method_description)}

---
"""
    )

    return remove_indent(doc)


def generate_help_md() -> str:

    doc = remove_indent(str(Mate.__doc__)) + "\n --- \n"
    current_docstring = str(sys.modules[__name__].__doc__)
    doc += current_docstring + "\n --- \n"
    members = [
        (k, v)
        for (k, v) in inspect.getmembers(Mate, predicate=inspect.isfunction)
        if not k.startswith("_")
    ]
    for name, val in members:
        doc += method_to_md(name, val)
    return doc


class MateHelp:
    def __init__(self):
        self.help_options = (
            "init",
            "install",
            "train",
            "auto",
            "export",
            "clone",
            "test",
        )
        self.help_comments = (
            "Initialize a new project",
            "Install dependencies",
            "Train a model",
            "Various commands to help with the development proces",
            "Export dependencies",
            "Clones internal modules",
            "Test a model",
        )

        self._full_docs, self._methods = self.get_full_docs(Mate)

        # save full docs to file
        # with open("docs.md", "w") as f:
        #     f.write(self._full_docs)

    def get_index(self):
        nl = "\n"
        # return self._full_docs

        helps = [
            f"`{option}` : {comment}"
            for option, comment in zip(self.help_options, self.help_comments)
        ]
        return remove_indent(
            f"""
        # Help Index

        Mate, your friendly ML project manager.
        
        Type `mate help <option>` to get more information about a topic.

        Available options are:

        {'- ' + (nl + ' - ').join(helps)}
        """
        )

    def get_full_docs(self, class_value):
        header = remove_indent(class_value.__doc__) + "\n\n --- \n"
        # gets the docstring of the methods and the variables with inspect
        def get_method_doc(method_name, method):
            doc = remove_indent(method.__doc__)
            signature = inspect.signature(method)
            args = list(signature.parameters.values())
            is_static = args[0].name != "self"
            args = [arg for arg in args if arg.name != "self"]
            annotations = ", ".join([str(a) for a in list(args)])
            header = f"### `mate` {method_name}\n"
            return header + doc + "\n --- \n"

        # methods = [
        #     get_method_doc(method_name, method)
        #     for method_name, method in inspect.getmembers(
        #         class_value, inspect.isfunction
        #     )
        #     if not method_name.startswith("_") and method.__doc__ is not None
        # ]

        method_dict = {
            method_name: get_method_doc(method_name=method_name, method=method)
            for method_name, method in inspect.getmembers(
                class_value, inspect.isfunction
            )
            if not method_name.startswith("_") and method.__doc__ is not None
        }

        nl = "\n"
        return (
            f"{header}\n\n{remove_indent(nl.join(method_dict.values()))}",
            method_dict,
        )

    def print_help(self, what: str = "") -> None:

        if what in self.help_options:
            # ipdb.set_trace()
            print_markdown(self._methods[what])
        else:
            print_markdown(self.get_index())
        # if what == "cli":
        #     print(generate_help_md())
        #     print_markdown(generate_help_md())
        # elif what == "mate":
        #     test = self.get_full_docs(Mate)
        #     print_markdown(test)
        # # elif what == "config":
        # #     print_markdown(MateConfig.__doc__)
        # # elif what == "project":
        # #     print_markdown(MateProject.__doc__)
        # # elif what == "experiment":
        # #     print_markdown(Experiment.__doc__)
        # else:
        #     # print_markdown(generate_help_md())
        #     print_markdown(self.get_index())

    def get_help_md(self, what: str = "cli") -> str:
        if what == "cli":
            return generate_help_md()
        elif what == "mate":
            return str(self.get_full_docs(Mate))
        # elif what == "config":
        #     return str(MateConfig.__doc__)
        # elif what == "project":
        #     return str(MateProject.__doc__)
        # elif what == "experiment":
        #     return str(Experiment.__doc__)
        else:
            return generate_help_md()

    def get_all_help_md(self) -> dict[str, str]:
        return {key: self.get_help_md(key) for key in self.help_options}


def collect_args(args: list[str], annotations: tuple[Callable]) -> tuple[list, dict]:
    # collects the arguments into a list and a dictionary
    # the list contains the positional arguments
    # the dictionary contains the keyword arguments
    # the arguments are split by the "=" sign
    # if there is no "=" sign, the argument is considered a positional argument
    # it also checks that there are no positional arguments after keyword arguments
    kwargs = {}
    positional_args = []
    positional_args_started = False

    def good_guess_type(arg: str):
        types = [int, float, str]
        if arg.lower() in ("none", "null"):
            return None
        elif arg.lower() == "true":
            return True
        elif arg.lower() == "false":
            return False
        for t in types:
            try:
                return t(arg)
            except ValueError:
                pass

    def boolean_type(arg: str):
        if arg.lower() == "true":
            return True
        elif arg.lower() == "false":
            return False
        else:
            raise ValueError("Not a boolean")

    if len(annotations) < len(args):
        annotations = annotations + (good_guess_type,) * (len(args) - len(annotations))
    from typing import Union

    for i, (arg, annotation) in enumerate(zip(args, annotations)):
        if annotation == bool:
            annotation = boolean_type
        elif annotation == inspect._empty:
            annotation = good_guess_type
        elif hasattr(annotation, "__args__"):
            annotation = annotation.__args__
        if "=" in arg:
            key, value = arg.split("=")
            kwargs[key] = annotation(value)
            positional_args_started = True
        else:
            if isinstance(annotation, tuple):
                annotation = annotation[0]
            positional_args.append(annotation(arg))
            if positional_args_started:
                raise ValueError(
                    f"positional argument {arg} after keyword argument {args[i-1]}"
                )

    return positional_args, kwargs


def main():
    methods = __get_methods_with_arguments(Mate)
    args = sys.argv[1:]
    raw_method_args = args[1:]
    help_args = ("help", "--help", "-h")
    actions = tuple(method.replace("_", "-") for method in methods) + help_args
    help = MateHelp()
    # ipdb.set_trace()
    if len(args) == 0 or args[0] in help_args:
        if len(args) > 1:
            help.print_help(args[1])
        else:
            help.print_help()
    else:
        # if args[0] not in actions:
        #     help.print_help()
        action = args[0]
        if len(args) > 1 and args[1] in ("--help", "-h"):
            md = method_to_md(action)
            print_markdown(md)

        elif action not in methods:
            try:
                mate = Mate()
                mate.train(*args[1:])
            except Exception as e:
                print(e)
                help.print_help()

        else:
            annotations = tuple(
                param.annotation
                for param in methods[action]
                if (param.kind != inspect.Parameter.VAR_KEYWORD)
                and (param.kind != inspect.Parameter.VAR_POSITIONAL)
            )

            pos_args, kwargs = collect_args(raw_method_args, annotations)
            if action == "init":
                Mate.init(*pos_args, **kwargs)
            else:
                mate = Mate()
                if hasattr(mate, action):
                    getattr(mate, action)(*pos_args, **kwargs)
                else:
                    if len(pos_args) == 2:
                        mate.train(*pos_args)



#!/usr/bin/python3
import re
import os
import urllib.request
import signal
import argparse
import json
import sys
from colorama import Fore, Style, init
import ipdb

init()

# this ANSI code lets us erase the current line
ERASE_LINE = "\x1b[2K"

COLOR_NAME_TO_CODE = {
    "default": "",
    "red": Fore.RED,
    "green": Style.BRIGHT + Fore.GREEN,
}


def print_text(
    text, color="default", in_place=False, **kwargs
):  # type: (str, str, bool, any) -> None
    """
    print text to console, a wrapper to built-in print

    :param text: text to print
    :param color: can be one of "red" or "green", or "default"
    :param in_place: whether to erase previous line and print in place
    :param kwargs: other keywords passed to built-in print
    """
    if in_place:
        print("\r" + ERASE_LINE, end="")
    print(COLOR_NAME_TO_CODE[color] + text + Style.RESET_ALL, **kwargs)


def create_url(url):
    """
    From the given url, produce a URL that is compatible with Github's REST API. Can handle blob or tree paths.
    """
    repo_only_url = re.compile(
        r"https:\/\/github\.com\/[a-z\d](?:[a-z\d]|-(?=[a-z\d])){0,38}\/[a-zA-Z0-9]+$"
    )
    re_branch = re.compile("/(tree|blob)/(.+?)/")

    # Check if the given url is a url to a GitHub repo. If it is, tell the
    # user to use 'git clone' to download it
    if re.match(repo_only_url, url):
        print_text(
            "✘ The given url is a complete repository. Use 'git clone' to download the repository",
            "red",
            in_place=True,
        )
        sys.exit()

    # extract the branch name from the given url (e.g master)
    branch = re_branch.search(url)
    download_dirs = url[branch.end() :]
    api_url = (
        url[: branch.start()].replace("github.com", "api.github.com/repos", 1)
        + "/contents/"
        + download_dirs
        + "?ref="
        + branch.group(2)
    )
    return api_url, download_dirs


def download(repo_url, output_dir="./"):
    """Downloads the files and directories in repo_url. If flatten is specified, the contents of any and all
    sub-directories will be pulled upwards into the root folder."""

    # generate the url which returns the JSON data
    api_url, download_dirs = create_url(repo_url)

    # ipdb.set_trace()

    # print("Output directory: " + output_dir)
    # print("Downloading from: " + api_url)
    # print("Download dirs: " + download_dirs)

    # ipdb.set_trace()

    dir_out = [d for d in download_dirs.split("/") if d != ""]
    dir_out = os.path.join(output_dir, *dir_out)
    # if len(download_dirs.split(".")) == 0:
    #     dir_out = os.path.join(output_dir, download_dirs)
    # else:
    #     dir_out = os.path.join(output_dir, *download_dirs.split("/")[:-1])

    # print(f"Downloading {repo_url} to: " + dir_out)

    try:
        opener = urllib.request.build_opener()
        opener.addheaders = [("User-agent", "Mozilla/5.0")]
        urllib.request.install_opener(opener)
        response = urllib.request.urlretrieve(api_url)
    except KeyboardInterrupt:
        # when CTRL+C is pressed during the execution of this script,
        # bring the cursor to the beginning, erase the current line, and dont make a new line
        print_text("✘ Got interrupted", "red", in_place=True)
        sys.exit()

    os.makedirs(dir_out, exist_ok=True)

    # total files count
    total_files = 0

    with open(response[0], "r") as f:
        data = json.load(f)
        # getting the total number of files so that we
        # can use it for the output information later
        total_files += len(data)

        # If the data is a file, download it as one.
        if isinstance(data, dict) and data["type"] == "file":
            try:
                # download the file
                opener = urllib.request.build_opener()
                opener.addheaders = [("User-agent", "Mozilla/5.0")]
                urllib.request.install_opener(opener)
                urllib.request.urlretrieve(
                    data["download_url"], os.path.join(dir_out, data["name"])
                )
                # bring the cursor to the beginning, erase the current line, and dont make a new line
                print_text(
                    "Downloaded: " + Fore.WHITE + "{}".format(data["name"]),
                    "green",
                    in_place=True,
                )

                return total_files
            except KeyboardInterrupt:
                # when CTRL+C is pressed during the execution of this script,
                # bring the cursor to the beginning, erase the current line, and dont make a new line
                print_text("✘ Got interrupted", "red", in_place=False)
                sys.exit()

        for file in data:
            file_url = file["download_url"]
            file_name = file["name"]
            file_path = file["path"]

            path = file_path
            dirname = os.path.dirname(path)

            # if dirname != "":
            #     os.makedirs(os.path.dirname(path), exist_ok=True)
            # else:
            #     pass

            if file_url is not None:
                try:
                    opener = urllib.request.build_opener()
                    opener.addheaders = [("User-agent", "Mozilla/5.0")]
                    urllib.request.install_opener(opener)
                    # download the file
                    urllib.request.urlretrieve(
                        file_url, os.path.join(dir_out, file_name)
                    )

                    # bring the cursor to the beginning, erase the current line, and dont make a new line
                    print_text(
                        "Downloaded: " + Fore.WHITE + "{}".format(file_name),
                        "green",
                        in_place=False,
                        end="\n",
                        flush=True,
                    )

                except KeyboardInterrupt:
                    # when CTRL+C is pressed during the execution of this script,
                    # bring the cursor to the beginning, erase the current line, and dont make a new line
                    print_text("✘ Got interrupted", "red", in_place=False)
                    sys.exit()
            else:
                download(file["html_url"], output_dir)

    return total_files


def main():
    if sys.platform != "win32":
        # disbale CTRL+Z
        signal.signal(signal.SIGTSTP, signal.SIG_IGN)

    parser = argparse.ArgumentParser(
        description="Download directories/folders from GitHub"
    )
    parser.add_argument(
        "urls", nargs="+", help="List of Github directories to download."
    )
    parser.add_argument(
        "--output_dir",
        "-d",
        dest="output_dir",
        default="./",
        help="All directories will be downloaded to the specified directory.",
    )

    parser.add_argument(
        "--flatten",
        "-f",
        action="store_true",
        help="Flatten directory structures. Do not create extra directory and download found files to"
        " output directory. (default to current directory if not specified)",
    )

    args = parser.parse_args()

    flatten = args.flatten
    for url in args.urls:
        total_files = download(url, flatten, args.output_dir)

    print_text("✔ Download complete", "green", in_place=True)


if __name__ == "__main__":
    main()



class DataSource:
    def __init__(self):
        pass

    def get_models(self, query: str = None):
        pass

    def get_trainers(self, query: str = None):
        pass

    def get_data_loaders(self, query: str = None):
        pass

    def get_experiments(self, query: str = None):
        pass

    def get_packages(self, query: str = None):
        pass



from enum import Enum
from typing import Any
import json
import ipdb


class Config:
    def __init__(self, config: dict):
        for key, value in self.__dict__.items():
            if value != None and value != {}:
                # Why assert when we can just set the default value?
                setattr(self, key, value)
                # assert key in config, f"Missing key {key} in config"
            if isinstance(value, Enum):
                enum_type = type(value)
                assert (
                    config[key] in enum_type.__members__.keys()
                ), "Invalid value for key {key}, must be one of {enum_type.__members__.keys()}"
                config[key] = type(value)(config[key])
            if key in config:
                assert isinstance(config[key], type(value)), f"Wrong type for key {key}"
                setattr(self, key, config[key])

        # Disable this for now as its changing the config
        # for key in config.keys():
        #     assert key in self.__dict__.keys(), f"Unknown key {key} in config."

    def __str__(self):
        return json.dumps(
            {
                key: (val if not isinstance(val, Metadata) else val.to_dict())
                for key, val in self.__dict__.items()
            },
            indent=4,
        )

    def __repr__(self):
        return self.__str__()

    def json(self):
        return {
            key: (val if not isinstance(val, Enum) else str(val))
            for key, val in self.__dict__.items()
        }

    def copy(self):
        return self.json().copy()

    # add mate["item"] = "value" functionality
    def __setitem__(self, key, value):
        setattr(self, key, value)

    # add mate["item"] functionality
    def __getitem__(self, key):
        return getattr(self, key)


class MateConfig(Config):
    def __init__(self, config):
        self.project = ""
        self.mate_version = ""
        self.results_folder = ""
        self.override_params: dict[str, Any] = {}
        self.restarting = False
        
        super().__init__(config)

    def save(self, path="mate.json"):

        with open(path, "w") as f:
            f.write(self.__str__())



from git import Repo
import os
import ipdb


def parse_url_from_git():

    max_depth = 3

    root_search = os.getcwd()
    nodes = []
    repo = None
    for i in range(max_depth):
        try:
            repo = Repo(root_search)
            break
        except:
            nodes.append(root_search.split("/")[-1])
            # ipdb.set_trace()
            root_search = os.path.join(root_search, "..")
            root_search = os.path.abspath(root_search)

    if repo:
        parsed_url = __parse_repo(repo)

        nodes = nodes[::-1]
        parsed_url = parsed_url + "/".join(nodes) if len(nodes) > 0 else parsed_url
        return parsed_url

    return None


def __parse_repo(repo):
    url = repo.remotes[0].url

    # transform url to https
    if url.startswith("git@"):
        # url = url.replace("git@", "https:/")
        # change git@github.com:oalee/deep-vision.git to https://github.com/oalee/deep-vision
        url = url.replace("git@", "https:/").replace(":", "/").replace("//", "://")

    # remove .git
    if url.endswith(".git"):
        url = url[:-4]

    branch = repo.active_branch.name
    url = f"{url}/tree/{branch}/"
    # ipdb.set_trace()
    return url




import json
import os
import shutil
import sys
import ipdb


ENV_KEY = "env"

class Environment(dict):

    def __init__(self):
        

        # default restart, test to False
        self.restart = False
        self.test = False
        self.train = False
        self.sample = False
        self.hparams = {}

        # set command to True
        setattr(self, sys.argv[1], True)
        

        self._root = self.__find_root()


        # parse python -m module train arg=1 arg2=2
        for arg in sys.argv:
            if "=" not in arg:
                continue
            key, value = arg.split("=")
            value = self.convert_str_to_data(value)
            setattr(self, key, value)
            self.hparams[key] = value

        
        self._path = sys.argv[0]
        if "bin/mate" in self._path:
            self.name = os.path.join(*sys.argv[2:4])
            self._path = os.path.join(self._root, "experiments", sys.argv[2],  sys.argv[3] + ".py")
            # detect path to experiment
        else:
            self.name = self._path.split("/")[-2:]
            self.name = os.path.join(*self.name)[: -3]

        self.__set_env()

        if self.train:
            self.__generate_experiment()

    def convert_str_to_data(self, input):
        try:
            return int(input)
        except ValueError:
            try:
                return float(input)
            except ValueError:
                if input in ["True", "true"]:
                    return True
                elif input in ["False", "false"]:
                    return False

        return input

    def __find_root(self):
        # TODO, find root of the project
        
        path = os.getcwd()

        if os.path.exists(os.path.join(path, "mate.json")):
            conf = json.load(open(os.path.join(path, "mate.json"), "r"))

            return os.path.join(path, conf["project"])
       
        return None

    def __generate_experiment(self):
        # copies the experiment to the results folder
        results =["results", "results_path", "results_dir", "save", "save_path", "save_dir"]
        result = None
        for key in results:
            if key in self.env:
                result = self.env[key]
                break
        if result is None:
            return
        if not os.path.exists(result):
            os.makedirs(result)
        
        save_file = os.path.join(result, "experiment.py")
        # copy from self._path to save_file
        shutil.copyfile(self._path, save_file)
       
        # ipdb.set_trace()

        if self.hparams != {}:
            params_file = os.path.join(result, "experiment.json")
            # dump self to params
            # overwrite if exists
            with open(params_file, "w") as f:
                json.dump(dict(self.hparams), f, indent=4)



    def __set_env(self):
        mate_conf_path = os.path.join("mate.json")
        conf = json.load(open(mate_conf_path, "r"))

        env_path = os.path.join("env.json")
        if not os.path.exists(env_path):
            print("Environment file not found, creating one with defaults: env.json")

            if ENV_KEY in conf:
                required_keys = conf[ENV_KEY].keys()
            else:
                required_keys = ["results"]

            defualt_env = {key: "" for key in required_keys}

            with open(env_path, "w") as f:
                json.dump(defualt_env, f, indent=4)

            
        with open(env_path, "r") as f:
            try:
                env = json.load(f)
            except json.decoder.JSONDecodeError:
                env = {
                    "results": "",
                }

        env_not_found = []

        if ENV_KEY in conf:
            for key, value in env.items():
                if env[key] == "":
                    if key in os.environ:
                        env[key] = os.environ[key]
                        # print(f"Environment variable {key} set to {os.environ[key]} from shell.")
                    else:
                        env_not_found.append(key)
                        env[key] = value
        
        if len(env_not_found) > 0:
            print(f"Environment variables not found. Set them in env.json or in your shell.")
            for key in env_not_found:
                desc = conf[ENV_KEY][key] if key in conf[ENV_KEY] else "Required environment variable"
                print(f"{key} : {desc}")

            print("Exiting...")
            sys.exit(1)
            
        # automatically append experiment name to results path
        search_results = ["results", "results_path", "results_dir", "save", "save_path", "save_dir"]
        for key in search_results:
            if key in env:
                env[key] = os.path.join(env[key], *self.name.split("/"))
                env["results"] = env[key]
                break
            if os.environ.get(key, None) is not None and os.environ.get(key, None) != "":
                env[key] = os.environ.get(key)
                env["results"] = env[key]
                break
            
        if len(env) == 0:
            print(f"results/save_dir/save environment variable is empty. Set it in env.json or in your shell.")
            print("Exiting...")
            sys.exit(1)
        
        setattr(self, "env", env)

    def __getitem__(self, key):
        if not key in self:
            # ipdb.set_trace()
            if key in self.env:
                return self.env[key]
            res = os.environ.get(key, None)

            if res is None or res == "":
                print(f"Environment variable {key} not found. Set it in env.json or in your shell.")
                print("Exiting...") 
                sys.exit(1)
            else:
                return res
            
        return super().__getitem__(key)

 


import os
import traceback
from yerbamate.api.data.sources.local import io
import json
from yerbamate.api.mate_api import MateAPI

import sys

import ipdb
from .utils import utils
import shutil
from typing import Optional

from .mate_config import MateConfig


class Mate:
    """
    Mate, your friendly ML project manager.
    """

    def __init__(self, init=False):
        self.root_folder = ""
        self.save_path = ""
        self.current_folder = os.path.dirname(__file__)
        self.config: Optional[MateConfig] = None
        self.__findroot()
        self.api = MateAPI(self.config)
        self.run_params = None

    @staticmethod
    def init(project_name: str, *args, **kwargs):
        """

        `mate init {project_name}`

        Initializes a new project.
        This will create the following structure:
        ```
        /
        |-- models/
        |   |-- __init__.py
        |-- experiments/
        |   |-- __init__.py
        |-- trainers/
        |   |-- __init__.py
        |-- data/
        |   |-- __init__.py
        ```

        Args:
            project_name: Name of the project.

        Example:

            `mate init my_project`

        """
        MateAPI.init_project(project_name, *args, **kwargs)

    def export(self):
        """
        Generates requirements.txt, dependencies.json and exports.md for sharing and reproducibility.

        Example:
            `mate export`

        Output:
            ```
            Generated requirements.txt for gan/models/lgan
            Generated dependencies.json for gan/experiments/lwgan
            Generated requirements.txt for gan/experiments/lwgan
            Generated requirements.txt for gan/trainers/lgan
            Generated requirements.txt for gan/data/cars
            Exported to export.md
            ```
        
        """
        self.api.auto("export")

    def list(
        self,
        module_name: str = None,
        output_json: bool = True,
    ):
        """


        Lists all available modules.

        Args:
            module_name: Name of the module to list. If not specified, all modules will be listed.

        Examples:
            `mate list models`

            `mate list`

        """

        li = self.api.list(module_name)
        if output_json:
            print(json.dumps(li, indent=4))
        else:
            print(li)

    def auto(self, command: str):
        """

        `mate auto {command}`

        Various commands to help with the development process.


        commands:
        - `export`: Creates a requirements.txt and dependencies.json files for sharing and reproducibility.
        - `init`: Automatically creates `__init__.py` files in the project structure.

        Example:


            `mate auto export`
            `mate auto init`
        """
        self.api.auto(command)

    def summary(self):
        """
        Prints a summary of the project modules. Same as `mate list`
        """

        print(json.dumps(self.api.summary(), indent=4))

    def clone(self, module: str, name: str, dest: str, *args, **kwargs):
        """
        `mate clone {module} {name} {dest} {-o}`

        Clones a source code module to a new destination.

        Args:

            module: Module to clone.

            name: Name of the module to clone.

            dest: Destination to clone the module to.

            -o: Overwrite destination if it exists.

        Example:

            `mate clone models torch_vit my_vit`
        """

        # check if module exists
        module_path = os.path.join(self.root_folder, module, name)
        if not os.path.exists(module_path):
            print(f"Module {module}/{name} does not exist")
            sys.exit(1)

        # check if destination exists
        dest_path = os.path.join(self.root_folder, module, dest)
        if os.path.exists(dest_path):

            if "-o" in args:
                shutil.rmtree(dest_path)
                print(f"Removed {dest_path}")
            else:
                print(f"Destination {dest} exists")
                overwrite = input("Overwrite? (y/n): ")
                if overwrite == "y":
                    shutil.rmtree(dest_path)
                    print(f"Removed {dest_path}")
                else:
                    sys.exit(1)

        # copy module to destination
        shutil.copytree(module_path, dest_path)

        print(f"Module {module}/{name} cloned to {dest_path}")

    def snapshot(self, model_name: str):
        pass

    def train(self, exp_module: str, exp: str, *args, **kwargs):
        """
        Executes an experiment.

        Usage: ``mate train {module} {experiment}``

        Args:

            exp_module: Name of the module where the experiment is located.

            exp: Name of the experiment.

        Example:
            `
            mate train experiments my_experiment`

        This will run the experiment `my_experiment` located in the `experiments` module.

        Equivalent to `python -m root_module.experiments.my_experiment train`
        """

        # ipdb.set_trace()
        module = [self.config.project, "experiments", exp_module, exp]
        module = ".".join(module)

        try:
            __import__(module)
        except Exception as e:
            print(f"Error in loading {exp_module}/{exp}")
            print("Available experiments:")
            self.list("experiments")
            traceback.print_exc()

            # print(e)
            sys.exit(1)

    def test(self, exp_module: str, exp: str, *args, **kwargs):
        """
        Executes an experiment.

        Usage: ``mate test {module} {experiment}``

        Args:

            exp_module: Name of the module where the experiment is located.

            exp: Name of the experiment.

        Example:
            `
            mate test experiments my_experiment`

        This will run the experiment `my_experiment` located in the `experiments` module.

        Equivalent to `python -m root_module.experiments.my_experiment test`
        """
        self.train(exp_module, exp, *args, **kwargs)

    def install(self, url: str, *args, **kwargs):
        """
        Intalls a module from a git repository.

        Usage: ``mate install {url} -{y|n|o} {pm}``

        Install module support the following formats:
        - ``mate install {complete_url}``
        - ``mate install {user}/{repo}/{root_module}/{module}``
        - ``mate install {user}/{repo|root_module}/{module}``

        Args:
        -    url: Url of the git repository.
        -    -y: Skips confirmation and installs python dependencies
        -    -n: Skips installing python dependencies
        -    -o: Overwrites existing code modules
        -    pm: Package manager to use. Defaults to asking the user.

        Example Installing a module from structured git repository (recommended):


            mate install oalee/deep-vision/deepnet/models/torch_vit -yo pip

            This will install the module `torch_vit` from the repository `oalee/deep-vision` in to your `models` folder.
            The `yo` flags will skip confirmation and install python dependencies using pip.

        Example Installing a module from unstructured git repository:


            mate install https://github.com/rwightman/pytorch-image-models/tree/main/timm


            This will install the module `timm` from the repository as a sister module to your root module.
            Take into account that this will install only the code and not the python dependencies.
        """
        self.api.install_url(url, *args, **kwargs)

    def __findroot(self):
        """
        Method in charge of finding the root folder of the project and reading the content of mate.json
        """
        self.root_folder, self.config = io.find_root()
        self.root_save_folder = self.config.results_folder



import os
import ipdb
import shutil

try:
    import validators
except ImportError as e:
    print("Please install validators module\n pip install validators")
    exit(1)


class GitUrlParser:
    def __init__(
        self, url: str = "https://github.com/ilex-paraguariensis/yerbamate/tree/v2/src"
    ):
        validators.url(url)
        self.base_url = url.split("tree")[0]
        self.branch = url.split("tree/")[1].split("/")[0]
        self.path = os.path.join(*url.split("tree/")[1].split("/")[1:])
        self.name = self.path.split("/")[-1]

    def clone(
        self, destination: str = "modules", tmp_destination: str = ".mate", force=False
    ):
        # os.makedirs(os.path.join(tmp_destination), exist_ok=True)
        destination = os.path.join(*destination.split("."))
        if os.path.exists(destination):
            if force:
                shutil.rmtree(destination)
            else:
                raise Exception(
                    f"Directory {destination} already exists. Please remove it and try again or use force=True"
                )
        if os.path.exists(tmp_destination):
            shutil.rmtree(tmp_destination)
        os.system(f"git clone -b {self.branch} {self.base_url} {tmp_destination}")
        shutil.rmtree(os.path.join(tmp_destination, ".git"))
        shutil.move(os.path.join(tmp_destination, self.path), destination)



import inspect
from typing import Callable
import ipdb
import os
from itertools import chain
import json

from .bunch import Bunch


def once(func):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return func(*args, **kwargs)

    wrapper.has_run = False
    return wrapper


print_once = once(print)


def get_function_parameters(
    function: Callable,
    object: dict,
    generate_defaults: bool = False,
    generate_none: bool = False,
):

    default = {}
    if "params" in object:
        params = object["params"].copy()
    else:
        params = {}

    error = None

    for param_name, param in inspect.signature(function).parameters.items():

        # only add parameters that are not self or exist in the params
        if param_name in params or param_name == "self":
            continue

        # only allow named parameters and not *args or **kwargs
        if param.kind != param.POSITIONAL_OR_KEYWORD:
            continue

        if param.default is not param.empty:
            # ignore None values
            if param.default != None and generate_defaults:
                default[param_name] = param.default
            elif generate_none:
                default[param_name] = None

        elif param.annotation is not param.empty:
            default[param_name] = f"Fix me! {param.annotation}"
            error = f"Missing parameter {param_name} for {object['module']}\nHint: {param.annotation}"
        else:
            default[param_name] = "Fix me!"
            error = f"Missing parameter {param_name} for {object['module']}\nHint: Add a default value or type annotation"

    params.update(default)

    return params, error


def migrate_mate_version(config: Bunch, root_project: str):

    from yerbamate import __version__
    from . import migrator

    if config.mate_version == __version__:  # pragma: no cover
        return config

    # handle migration from current_version to __version__
    migrator = migrator.Migration(root_project, config.mate_version, __version__)

    success = migrator.migrate()
    if not success:
        raise Exception(
            "Migration failed, please send the logs and open an issue on github"
        )

    config.mate_version = __version__

    with open(os.path.join(root_project, "mate.json"), "w") as f:
        json.dump(config, f, indent=4)

    print("Migration sucessfull!")
    return Bunch(config)

    # if self.config.mate_version != __version__:
    #     print(
    #         f"New mate version has been installed. Going to handle migration from {self.config.mate_version} to {__version__}"
    #     )
    #     migrator = Migration(
    #         self.root_folder, self.config.mate_version, __version__
    #     )
    #     success = migrator.migrate()
    #     if not success:
    #         print(
    #             "Migration failed... please check the logs and make an issue on github"
    #         )
    #         # sys.exit(1)
    #     else:
    #         self.config.mate_version = __version__
    #         with open(path, "w") as f:
    #             json.dump(self.config, f, indent=4)
    #         print("Migration successful")

    pass


def get_model_parameters(path: str):
    path = path.replace("/", ".")
    model_class = __import__(
        f"{path}.model",
        fromlist=[path.split(".")[1]],
    ).Model
    return get_function_parameters(model_class.__init__)


def is_leaf(path: str):
    children = [u for u in [p for p in os.listdir(path)] if os.path.isdir(u)]
    return os.path.exists(os.path.join(path, "results")) or len(children) == 0


def get_leaves_rec(path: str):
    sub = [os.path.join(path, p) for p in os.listdir(path)]
    dirs = list(
        chain([get_leaves_rec(s)[0] for s in sub if os.path.isdir(s) if not "__" in s])
    )
    return dirs if not is_leaf(path) else [path]


def get_parameters(path: str, dataset: str = "default"):
    # leaves = get_leaves_rec(path)
    config = os.path.join(path, "parameters", f"{dataset}.json")
    if os.path.exists(config):
        with open(config) as f:
            result = json.load(f)
        return result
    else:
        return get_model_parameters(path)

    # result = [
    #    (leaf.replace("/", "."), get_model_parameters(leaf)) for leaf in leaves
    # ]
    # return result


if __name__ == "__main__":
    print(get_parameters("ciao"))



import json
import os
import shutil
import sys
from typing import Optional
import ipdb

from .....mate_config import MateConfig
from .....utils.bunch import Bunch
from .....utils.utils import once



print_once = once(print)


def find_root():
    """
    Method in charge of finding the root folder of the project and reading the content of mate.json
    """
    # path of execution
    current_path = os.getcwd()
    found = False
    i = 0
    root_folder = ""
    config: Optional[MateConfig] = None
    while not found and i < 2:
        if os.path.exists(os.path.join(current_path, "mate.json")):
            conf_path = os.path.join(current_path, "mate.json")
            config = load_mate_config(conf_path)
            root_folder = config.project
            # self.__import_submodules(self.root_folder)
            found = True
        else:
            os.chdir("..")
            current_path = os.getcwd()
            i += 1
            if current_path == "/" or i == 2:
                raise Exception(
                    "Could not find mate.json. Please make sure you are in the root folder of the project."
                )

    # self.root_save_folder = self.root_folder
    sys.path.insert(0, os.getcwd())
    return root_folder, config



def list_packages(root_folder: str, folder: str):
    return (
        tuple(x for x in os.listdir(os.path.join(root_folder, folder)) if not "__" in x)
        if os.path.exists(os.path.join(root_folder, folder))
        else tuple()
    )


def get_experiment_description(root_folder: str, model_name: str, experiment: str):
    # ipdb.set_trace()
    experiments = list_experiments(root_folder, model_name, False)
    for exp in experiments:
        if exp[1] == experiment and exp[2] == model_name:
            return exp
        if exp[1] == model_name and exp[2] == "experiments":
            return exp
    return None


def list_experiments(root_foolder: str, model_name=None, log=True):
    """
    models = list_packages(root_foolder, "models")

    dirs = [
        (os.path.join(root_foolder, "experiments"), "experiments"),
    ]

    if model_name != None and model_name in models:
        models = [model_name]
        dirs = []

    for model in models:
        dirs.append((os.path.join(root_foolder, "models", model, "experiments"), model))

    experiments = []

    for dir, model in dirs:
        if os.path.exists(dir):
            files = os.listdir(dir)
            for file in files:
                if not "__" in file and ".json" in file:
                    experiments.append([dir, file.replace(".json", ""), model])

    if log:
        for (
            dir,
            param_file,
            model,
        ) in experiments:
            print(f"{model}: {dir} {param_file}")

    paths = [x[0] for x in experiments]
    names = [x[1] for x in experiments]
    models = [x[2] for x in experiments]

    return experiments
    """
    return os.listdir(os.path.join(root_foolder, "experiments"))


def list_experiment_names(root_folder: str, model_name: str):
    experiments = list_experiments(root_folder, None, False)
    return [x[1] for x in experiments]


# def experiment_exists(root_folder: str, model_name: str, experiment_name: str):
#     # ipdb.set_trace()
#     exp_path = __get_experiment_path(root_folder, model_name, experiment_name)
#     return os.path.exists(exp_path)


# def assert_experiment_exists(root_folder: str, model_name: str, experiment_name: str):
#     # ipdb.set_trace()
#     exp_path = __get_experiment_path(root_folder, model_name, experiment_name)

#     assert os.path.exists(exp_path), f"Experiment {exp_path} does not exist"


def load_mate_config(path):
    with open(path) as f:
        config = MateConfig(json.load(f))
        # assert (
        #    "results_folder" in config
        # ), 'Please add "results_folder":<path> in mate.json'
        # assert "project" in config, 'Please add "project":<project name> in mate.json'
    return config


def remove(root_folder: str, model_name: str):
    action = "go"
    while action not in ("y", "n"):
        action = input(f'Are you sure you want to remove model "{model_name}"? (y/n)\n')
    if action == "y":
        shutil.rmtree(os.path.join(root_folder, "models", model_name))
        print(f"Removed model {model_name}")
    else:
        print("Ok, exiting.")


def list(root_folder: str, folder: str):
    print("\n".join(tuple("\t" + str(m) for m in list_packages(root_folder, folder))))


def clone(root_folder: str, source_model: str, target_model: str):
    shutil.copytree(
        os.path.join(root_folder, "models", source_model),
        os.path.join(root_folder, "models", target_model),
    )


def snapshot(root_folder: str, model_name: str):

    if not os.path.exists(os.path.join(root_folder, "snapshots")):
        os.makedirs(os.path.join(root_folder, "snapshots"))

    snapshot_names = [
        name.split("__") for name in os.listdir(os.path.join(root_folder, "snapshots"))
    ]
    matching_snapshots = [name for name in snapshot_names if name[0] == model_name]
    max_version_matching = (
        max([int(name[1]) for name in matching_snapshots])
        if len(matching_snapshots) > 0
        else 0
    )
    snapshot_name = f"{model_name}__{max_version_matching + 1}"

    shutil.copytree(
        os.path.join(root_folder, "models", model_name),
        os.path.join(root_folder, "snapshots", snapshot_name),
    )
    print(f"Created snapshot {snapshot_name}")



import json
import sys
from yerbamate.mate_config import MateConfig
from yerbamate.utils.bunch import Bunch
from ..source import DataSource
from glob import glob
import os
from typing import Optional
import ipdb
from ..local import io
from typing import Any


class LocalDataSource(DataSource):
    def __init__(
        self,
        config: MateConfig,
        # root_dir: str = ".",  # a different root (".mate") is used while installing a packaage installing a package, we can use the root dir of the package
    ):
        super().__init__()

        root_dir = config.project
        self.root_folder = root_dir
        self.config = config
        self.map = None

        # ProjectParser.check_project_structure(root_dir)

        self.__load_data(root_dir)

    def load_mate_config_and_root(self):
        return self.__findroot()

    def summary(self):
        self.__load_data(self.root_folder)
        return self.map

    def __load_data(self, root_dir: str):

        if self.map != None:
            return

        self.map = {}
        for dir in os.listdir(root_dir):

            if dir == "experiments":
                self.map["experiments"] = {
                    dir: [
                        exp_file
                        for exp_file in os.listdir(
                            os.path.join(root_dir, "experiments", dir)
                        )
                        if ".py" in exp_file and exp_file != "__init__.py"
                    ]
                    for dir in os.listdir(os.path.join(root_dir, "experiments"))
                    if os.path.isdir(os.path.join(root_dir, "experiments", dir))
                    and dir != "__pycache__"
                    or (".py" in dir and dir != "__init__.py")
                }
                # self.map["experiments"] =  exps.values
            else:
                if os.path.isdir(os.path.join(root_dir, dir)) and dir != "__pycache__":
                    # see if there is a __init__.py file
                    if "__init__.py" in os.listdir(os.path.join(root_dir, dir)):
                        self.map[dir] = [
                            sub_dir
                            for sub_dir in os.listdir(os.path.join(root_dir, dir))
                            if os.path.isdir(os.path.join(root_dir, dir, sub_dir))
                            and os.path.exists(
                                os.path.join(root_dir, dir, sub_dir, "__init__.py")
                            )
                            and sub_dir != "__pycache__"
                        ]
        # ipdb.set_trace()
        # self.map = {k: v for k, v in self.map.items()}

    def __filter_regular_folders(self, names: list[str]):
        return [fn for fn in names if not fn.startswith("__")]

    def __filter_names(self, query: Optional[str], names: list[str]):
        return names if query is None else [name for name in names if query == name]

    def list(self, module: str):

        if module is None:
            return self.map
        # ipdb.set_trace()
        assert module in self.map.keys(), f"Module {module} not found"
        return self.map[module]

    def __findroot(self):
        """
        Method in charge of finding the root folder of the project and reading the content of mate.json
        """
        self.root_folder, self.config = io.find_root()
        self.root_save_folder = self.config.results_folder
        self.save_path = self.root_save_folder
        return self.config, self.root_folder, self.root_save_folder

    def __validate_missing_params(
        self,
        root: Bunch,
        model_name: str,
        experiment: str,
        generate_defaults: bool = False,
    ):
        """
        Validates that all the required parameters are present in the params file
        """

        parsed_params, errors = self.trainer.generate_full_dict()

        if len(errors) > 0:
            print(f"Errors in {model_name}/{experiment}")
            for error in errors:
                print(error)

            # io.update_hyperparameters(
            #     self.root_folder, model_name, experiment, parsed_params
            # )
            sys.exit(1)

        return parsed_params

    def __parse_and_validate_params(self, trainer, experiment: str):
        assert (
            self.trainer is not None
        ), "Trainer must be initialized before parsing params (Bombilla is managed by Trainer)"

        full, err = trainer.generate_full_dict()

        if len(err) > 0:
            print(f"Errors in {experiment}")
            for error in err:
                print(error)

            # io.update_hyperparameters(self.root_folder, model_name, params, full)
            sys.exit(1)

        io.save_train_experiments(self.save_path, full, self.config)

        return full

    def add_model(self, model):
        self.models.append(model)

    def add_trainer(self, trainer):
        self.trainers.append(trainer)

    def add_data_loader(self, dataset):
        self.data_loaders.append(dataset)

    def add_experiment(self, experiment):
        self.experiments.append(experiment)

    def add_package(self, package):
        self.packages.append(package)



from rich.padding import Padding
from rich.console import Console
from rich.markdown import Markdown
from glob import glob
import os
import shutil
import ipdb

# remove_indent = lambda text: "\n".join(
#     [line[len(line) - len(line.lstrip()) :] for line in text.split("\n")]
# )
def remove_indent(text: str):
    indents: dict[int, int] = {}
    for line in [l for l in text.splitlines() if l.strip() != ""]:
        indent = len(line) - len(line.lstrip())
        if indent not in indents:
            indents[indent] = 0
        indents[indent] += 1

    selected_indent = max(indents.items(), key=lambda x: x[1])[0]
    def indent_remover(line: str):
        current_indent = len(line) - len(line.lstrip())
        indent = min(current_indent, selected_indent)
        return line[indent:]

    return "\n".join([indent_remover(line) for line in text.split("\n")])


console = Console(width=75)
print = console.print
def print_markdown(x, svg:bool=False, file_name:str="", title:str=""):
    if svg:
        console = Console(width=50, record=True)
        local_print = console.print
    else:
        local_print = print
    local_print(Padding(Markdown(remove_indent(x)), (1, 0, 0, 5)))
    if svg:
        console.save_svg(file_name, title=title)
        


def rmwithin(path: str):
    for file in glob(path + "*"):
        if os.path.isdir(file):
            shutil.rmtree(file)
        else:
            os.remove(file)



from ..source import DataSource


# TODO: this class should call API to get the data
class RemoteDataSource(DataSource):
    def __init__(self):
        super().__init__()
        self.models = []
        self.trainers = []
        self.datasets = []
        self.experiments = []
        self.packages = []

    def get_models(self, query: str = None):
        return self.models

    def get_trainers(self, query: str = None):
        return self.trainers

    def get_data_loaders(self, query: str = None):
        return self.datasets

    def get_experiments(self, query: str = None):
        return self.experiments

    def get_packages(self, query: str = None):
        return self.packages

    def add_model(self, model):
        self.models.append(model)

    def add_trainer(self, trainer):
        self.trainers.append(trainer)

    def add_dataset(self, dataset):
        self.datasets.append(dataset)

    def add_experiment(self, experiment):
        self.experiments.append(experiment)

    def add_package(self, package):
        self.packages.append(package)



from types import SimpleNamespace

import json
import ipdb


class Bunch(dict):
    def __init__(self, kwargs):
        super().__init__(kwargs)
        for key, val in kwargs.items():
            self.__dict__[key] = val if not isinstance(val, dict) else Bunch(val)

    def to_dict(self):
        proto = self.__dict__.copy()
        for key, val in proto.items():
            proto[key] = val if not isinstance(val, Bunch) else val.to_dict()
        return proto

    def contains(self, key):
        return key in self.keys()

    def has(self, key):
        return self.contains(key)

    def is_empty(self):
        return len(self.keys()) == 0

    def clone(self):
        return Bunch(self.to_dict())

    def __str__(self):
        return json.dumps(self.to_dict(), indent=4, default=str)

    def __setattr__(self, key, value):
        self.__setitem__(key, value)

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        return self.__getitem__(key)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self.__dict__[key] = value

    def __getitem__(self, key):
        return super().__getitem__(key)

    def __setstate__(self, state):
        # Bunch pickles generated with scikit-learn 0.16.* have an non
        # empty __dict__. This causes a surprising behaviour when
        # loading these pickles scikit-learn 0.17: reading bunch.key
        # uses __dict__ but assigning to bunch.key use __setattr__ and
        # only changes bunch['key']. More details can be found at:
        # https://github.com/scikit-learn/scikit-learn/issues/6196.
        # Overriding __setstate__ to be a noop has the effect of
        # ignoring the pickled __dict__
        pass



class Constants:
    class File:
        HYPERPARAMETER_FOLDER = "hyperparameters"
        MODELS_FOLDER = "models"
        MATE_CONFIG = "mate.json"
        DATA_LOADER_FOLDER = "data_loaders"
        DATA_LOADER_FILE = "data_loader"
        EXEC_FOLDER = "exec"
        EXEC_FUNCTION = "run"



from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name="yerbamate",
    description=" A python module and experiment manager for deep learning",
    author="Giulio Zani, Ali Rahimi",
    author_email="yerba.mate.dl@proton.me",
    url="https://github.com/oalee/yerbamate",
    python_requires=">=3.9",
    version="0.9.21",
    packages=find_packages("packages", exclude=["tests"]),
    include_package_data=True,
    package_dir={"": "packages/"},
    license="Apache License 2.0",
    license_files=("LICENSE",),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "colorama",
        "GitPython",
        "validators",
        "ipdb",
        "pipreqs",
        "dirhash",
        "rich",
        "docstring-parser",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3"
    ],
    scripts=["./src/mate"],
)



import ast

def get_relative_imports(file):
    with open(file, 'r') as f:
        tree = ast.parse(f.read())
    relative_imports = []
    for node in ast.walk(tree):
        if isinstance(node, ast.ImportFrom):
            if node.level > 0:
                relative_imports.append(node.module)
    return relative_imports




# A Mate package, with type, name, source, version, and dependencies
# This is the base class for all packages.
# A package could be a Model, Trainer, DataModule and so on.
# A package needs to have a install method, which will be called by the package manager?
# License: GPL v3

import os
import sys
import shutil
import urllib
from typing import Union, Callable, Optional, Type
from ...utils.bunch import Bunch

import inspect
import ipdb


class Package:

    # Type could become Enum, for now it's a string e.g. "model", "trainer"
    # Source could be a URL, or a local path?
    # Destination could be a local path, for example "models", "trainers", "data", "models/vit", "models/cnn/resnet"
    def __init__(
        self,
        params: Bunch = None,
        root: str = None,
        backbone: str = None,
        source: str = None,
        package: str = None,
        type: str = None,
        url: str = None,
        export: Bunch = None,
        description: str = None,
        version: str = None,
        author: str = None,
        license: str = None,
        pip: Bunch = None,
    ):
        self.type = type
        self.source = source  # find a way to read the source
        self.root = root
        self.backbone = backbone
        self.url = url
        self.params = params

        self.export = export
        self.package = package

        self.description = description
        self.version = version
        self.author = author
        self.type = type
        # self.object = self._object()

    @staticmethod
    def install(source: str, destination: str):
        pass

    def _get_package_class(self, package_location: str, package_name: str) -> dict:
        # package_name = package_location.split(".")[-1]
        # package_class_name = package_name.title().replace("_", "")
        package = __import__(
            package_location,
            fromlist=[package_location.split(".")[1]],
        )
        package_class = getattr(package, package_name)
        ipdb.set_trace()
        return package_class

    def _generate_signature(self, entrypoint: Union[Type, Callable]) -> dict:
        return inspect.getfullargspec(entrypoint).args  # TODO: turn it into a dict

    def _parse_signature(self, args: dict) -> object:
        return self._get_package_class(self.root, self.source)(**args)

    def _object(self) -> object:
        # return self._parse_signature(self.params)
        return self._get_package_class(self.root, self.source)

    def update(self) -> None:
        pass

    def uninstall(self) -> None:
        pass

    def __str__(self):
        return f"Package: {self.type} {self.source}"


class PLTrainerPackage(Package):

    # TODO automatically parse root, package, version from mate config

    # only params are required
    def __init__(
        self,
        params: Bunch = None,
        root: str = None,
        backbone: str = "torch",
        source: str = "local",
        package: str = None,
        type: str = "pl_train",
        url: str = None,
        export: Bunch = None,
        description: str = None,
        version: str = None,
    ):
        super().__init__(
            root=root,
            backbone=backbone,
            source=source,
            package=package,
            type=type,
            url=url,
            params=params,
            export=export,
            description=description,
            version=version,
        )



import os

import sys

from time import sleep
from yerbamate.mate_config import MateConfig


from .data.module_repository import ModuleRepository

import ipdb


"""
MATE API

"""


class MateAPI:
    def __init__(self, config: MateConfig):
        self.config: MateConfig = config
        self.repository = ModuleRepository(config)


    @staticmethod
    def init_project(project_name: str):
        ModuleRepository.init_project(project_name)

    # def generate_metadata(self, rewrite: bool = False):
    #     return self.repository.metadata_generator.generate(rewrite)

    def install_url(self, url: str, *args, **kwargs):
        self.repository.install_url(url, *args, **kwargs)

    def list(self, module: str):
        return self.repository.list(module)
    

    def summary(self):
        return self.repository.get_mate_summary()

    def auto(self, command: str):
        self.repository.auto(command)









import hashlib
import json
import os
import re
from shutil import copytree, rmtree
import validators
from yerbamate.mate_config import MateConfig

from .utils.gitdir import download
import ipdb


class ModuleManager:
    root_path: str = ""

    def __init__(self, conf: MateConfig):
        self.conf = conf

        self.init_package_cache()

    def init_package_cache(self):

        folder_name = ".mate"
        if not os.path.exists(folder_name):
            os.mkdir(folder_name)

        self.root_path = os.path.join(os.getcwd(), folder_name)

    def get_path(self, *args):
        return os.path.join(self.root_path, *args)

    def check_auto_install_reqs(self, *args, **kwargs):
        if len(args) < 1:
            return False
        if args[0] == "-y" or "y" in args[0]:
            return True

    def check_auto_no_install_reqs(self, *args, **kwargs):
        if len(args) < 1:
            return False
        if args[0] == "-n" or "n" in args[0]:
            return True

    def auto_install_manager(self, *args, **kwargs):
        if self.check_auto_install_reqs(*args, **kwargs):
            if len(args) > 1:
                pm = args[1]
                if pm in ["pip", "conda"]:
                    return pm
            else:
                cmd = input("Install with pip or conda? [pip/conda]: ")
                if cmd in ["pip", "conda"]:
                    return cmd

            while cmd not in ["pip", "conda"]:
                cmd = cmd("Install with pip or conda? [pip/conda]: ")
                if cmd in ["pip", "conda"]:
                    return cmd

            return None
        return None

    def check_auto_overwrite(self, *margs, **kwargs):
        if len(margs) < 1:
            return False
        # ipdb.set_trace()

        try:
            args = margs[0][0]

            for arg in args:
                if re.match("-\w*o", arg):
                    return True
        except:
            return False

        # for arg in margs:
        #     if type(arg) is tuple:
        #         for a in list(arg):
        #             if type(a) is str:
        #                 if re.match("-[*]?o[*]?", a):
        #                     return True

        #     else:
        #         regex = r"-[*]?o[*]?"
        #         if type(arg) is str:
        #             if re.match(regex, arg):
        #                 return True
        # return False

    def install_package(self, url: str, *args, **kwargs):

        default_trees = ["main", "master"]

        # ipdb.set_trace()
        if url.count("/") > 3 and "https" not in url:

            url = url.split("/")
            # oalee/deep-vision/deepnet/models/resnet to https://github.com/oalee/deep-vision/tree/main/deepnet/models/resnet

            url = f"https://github.com/{url[0]}/{url[1]}/tree/main/{'/'.join(url[2:])}"
            # if not validators.url(url):
            #     url = f"https://github.com/{url[0]}/{url[1]}/tree/master/{'/'.join(url[2:])}"

        elif url.count("/") == 3 and "https" not in url:

            # same repo and module name
            # oalee/big_transfer/experiments/bit
            url = url.split("/")
            url = f"https://github.com/{url[0]}/{url[1]}/tree/main/{'/'.join(url[1:])}"

            # if not validators.url(url):
            #     url = f"https://github.com/{url[0]}/{url[1]}/tree/master/{'/'.join(url[1:])}"
            # # ipdb.set_trace()

        assert validators.url(url), "Invalid url"
        package_install_dst = self.__install_package(url, args, kwargs)

        # check for requirements.txt
        if package_install_dst:
            requirements_path = os.path.join(package_install_dst, "requirements.txt")
            if os.path.exists(requirements_path):

                if self.check_auto_install_reqs(*args, **kwargs):
                    pm = self.auto_install_manager(*args, **kwargs)
                    if pm == "pip":
                        os.system(f"pip install -r {requirements_path}")
                        print("")
                    elif pm == "conda":
                        os.system(f"conda install --file {requirements_path}")
                        print("")
                    else:
                        print("Skipping requirements installation")

                elif not self.check_auto_no_install_reqs(*args, **kwargs):
                    cmd = input(
                        "Requirements found. Do you want to install them with pip? [y/conda/n]: "
                    )
                    if cmd == "y":
                        os.system(f"pip install -r {requirements_path}")
                        print("")

                    elif cmd == "conda":
                        os.system(f"conda install --file {requirements_path}")
                        print("")
                    else:
                        print("Skipping requirements installation")
            else:
                print("No requirements found. Manually check and install dependencies.")
            # read dependencies.json
            dependencies_path = os.path.join(package_install_dst, "dependencies.json")
            if os.path.exists(dependencies_path):
                with open(dependencies_path, "r") as f:
                    dependencies = json.load(f)
                for dependency in dependencies["dependencies"]:
                    # ipdb.set_trace()
                    self.install_package(dependency, *args, **kwargs)

        # update history

        print(f"Module installed at {package_install_dst}")

    def __is_valid_module(self, name):
        # module names, such as models.my_model, or models or my_model.e.z
        return re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*$", name)
        # return re.match(r"^[a-zA-Z_][a-zA-Z0-9_]*$", name)

    def __install_package(self, url, *args, **kwargs):

        # just git url
        base_git_url = "/".join(url.split("/")[:7])

        url_hash = hashlib.sha1(base_git_url.encode("utf-8")).hexdigest()

        output_path = self.get_path(url_hash)
        output_dir = output_path

        if not os.path.exists(output_path):
            os.makedirs(output_path, exist_ok=True)

        root_module = url.split("/")[-2]
        module_name = url.split("/")[-1]

        print(f"Installing {root_module}/{module_name} from {base_git_url}")

        if root_module in ["models", "data", "experiments", "trainers"]:
            dest_path = os.path.join(
                os.getcwd(), self.conf.project, root_module, module_name
            )
            # command = input(f"Install {self.conf.project}/{root_module}/{module_name} at {dest_path} [y,n]? ")
        else:
            # ipdb.set_trace()
            if module_name == "":
                dest_path = os.path.join(os.getcwd(), root_module)
            elif root_module in ["master", "main"]:
                dest_path = os.path.join(os.getcwd(), module_name)
            else:
                print(
                    "Could not automatically determine the module type. Please specify."
                )
                command = input(
                    f"Is {self.conf.project}.{root_module}.{module_name} Correct [y,n]? "
                )
                if command == "y":
                    dest_path = os.path.join(
                        os.getcwd(), self.conf.project, root_module, module_name
                    )
                else:
                    dest_path = input(
                        "Please specify the correct module: (e.g. models.my_model) "
                    )
                    while not self.__is_valid_module(dest_path):
                        dest_path = input(
                            "Please specify the correct module: (e.g. models.my_model) "
                        )
                    dest_path = dest_path.split(".")
                    dest_path = os.path.join(os.getcwd(), self.conf.project, *dest_path)

        while os.path.exists(dest_path):
            if not self.check_auto_overwrite(args, kwargs):
                cmd = input(
                    "Package already exists, options: [o]verwrite, [c]ancel, [r]ename: "
                )
            else:
                cmd = "o"

            if cmd == "o":
                rmtree(dest_path)
                print("Deleted existing package at ", dest_path)
                break

            elif cmd == "c":
                return

            elif cmd == "r":
                dest_path = input(
                    "Please specify the correct module: (e.g. models.my_model) "
                )
                while not self.__is_valid_module(dest_path):
                    dest_path = input(
                        "Please specify the correct module: (e.g. models.my_model) "
                    )
                dest_path = dest_path.split(".")
                dest_path = os.path.join(os.getcwd(), self.conf.project, *dest_path)
                break

        print(f"Downloading {url}")

        try:
            download(url, output_dir=output_dir)
        except:
            # try master branch
            url = url.replace("/main/", "/master/")
            print(f"Downloading {url}")
            download(url, output_dir=output_dir)

        src_module = url.split("tree")[1].split("/")[2:]
        src_module = os.path.join(output_dir, *src_module)

        # if "timm" in dest_path:
        #     ipdb.set_trace()

        copytree(src_module, dest_path)
        rmtree(output_dir)
        return dest_path
